#include "common.h"

#include "ximc.h"
#include "util.h"
#include "protosup.h"
#include "loader.h"
#include "ximc-gen.h"
#include "platform.h"

/*
 * Types
 */

/* Opaque struct with the enumerate devices information */
typedef struct device_enumeration_opaque_t
{
	int allocated_count;
	int count;
	int flags;
	char** names;
	uint32_t* serials;
	device_information_t* infos;
} device_enumeration_opaque_t;

typedef struct thread_state_t
{
	char* name;
	uint32_t serial;
	device_information_t* info;
	int status;
} thread_state_t;


/*
 * Device enumeration implementation
 */

/* doesn't lock, serial is optional
*/
int check_device_by_ximc_information (const char* name, device_information_t* info, uint32_t* serial)
{
	device_t device;
	device_information_t info_local;
	device_information_t* pinfo = info ? info : &info_local;
	int is_ximc_device = 0;

	log_debug( L"enum thread: started, %hs", name );

	device = open_device_impl( name, ENUMERATE_TIMEOUT_TIME );
	if (device != device_undefined)
	{
		log_debug( L"enum thread: opened, starting GETI" );
		if (get_device_information_impl_unsynced( device, pinfo ) == result_ok)
		{
			log_debug( L"enum thread: successfully got GETI" );
			if (!strcmp( pinfo->Manufacturer, "XIMC" ))
			{
				is_ximc_device = 1;
				if (serial)
				{
					log_debug( L"enum thread: starting GSER" );
					if (get_serial_number( device, serial ) != result_ok)
					{
						log_warning( L"Cannot get serial number from device %hs", name );
						*serial = 0;
					}
				}
			}
		}
		log_debug( L"enum thread: closing" );
		close_device_impl( &device );
	}
	return is_ximc_device;
}

/* Concrete thread function to slowly check a device for beeing XIMC */
void check_device_thread (void* arg)
{
	thread_state_t* ts = (thread_state_t*)arg;
	ts->status = check_device_by_ximc_information( ts->name, ts->info, &ts->serial ) ? 1 : 0;
}

/*
 * Launch device check threads
 * Remove non-passed device names from devenum
 */
void launch_check_threads(device_enumeration_opaque_t* devenum)
{
	/* Check all found devices in threads */
	thread_state_t* tstates;
	int i, k;

	log_debug( L"precheck found %d devices, launching check threads", devenum->count );

	tstates = (thread_state_t*)malloc( devenum->count*sizeof(thread_state_t) );

	for (i = 0; i < devenum->count; ++i)
	{
		log_info( L"launch thread to probe device %hs", devenum->names[i] );
		tstates[i].name = devenum->names[i];
		tstates[i].info = &devenum->infos[i];
	}

	if (fork_join( check_device_thread, devenum->count, tstates, sizeof(thread_state_t) ) != result_ok)
	{
		log_error( L"fork/join engine failed" );
		devenum->count = 0;
	}

	for (i = 0, k = 0; i < devenum->count; ++i)
	{
		log_debug( L"Check device %hs finished with %d", devenum->names[i], tstates[i].status );
		/* rewrite devenum array */
		if (tstates[i].status)
		{
			devenum->names[k] = tstates[i].name;
			devenum->serials[k] = tstates[i].serial;
			if (&devenum->infos[k] != tstates[i].info)
				memcpy( &devenum->infos[k], tstates[i].info, sizeof(device_information_t) );
			++k;
		}
	}
	devenum->count = k;
	free( tstates );
}

/* Concrete callback function that saves provided device name into enumerator */
void store_device_name (char* name, void* arg)
{
	device_enumeration_opaque_t* devenum = (device_enumeration_opaque_t*)arg;
	int i;

	for (i = 0; i < devenum->count; ++i)
	{
		if (is_same_device( name, devenum->names[i] ))
		{
			log_debug( L"Skipping duplicate device %hs", name );
			return;
		}
	}

	log_debug( L"Storing port %hs", name );

	if (devenum->count >= devenum->allocated_count)
	{
		devenum->allocated_count = (int)(devenum->allocated_count * 1.5);
		devenum->names = (char**)realloc( devenum->names, devenum->allocated_count*sizeof(char*) );
		devenum->serials = (uint32_t*)realloc( devenum->serials, devenum->allocated_count*sizeof(uint32_t) );
		devenum->infos = (device_information_t*)realloc( devenum->infos, devenum->allocated_count*sizeof(device_information_t) );
	}

	devenum->names[devenum->count] = portable_strdup( name );
	++devenum->count;
}

/* Enumerate devices main function */
result_t enumerate_devices_impl(device_enumeration_opaque_t** device_enumeration, int enumerate_flags)
{
	device_enumeration_opaque_t* devenum;

	/* alloc mem */
	*device_enumeration = (device_enumeration_opaque_t*)malloc(sizeof(device_enumeration_opaque_t));
	devenum = *device_enumeration;
	devenum->flags = enumerate_flags;
	devenum->count = 0;
	devenum->allocated_count = 16;
	devenum->names = (char**)malloc( devenum->allocated_count * sizeof(char*) );
	devenum->serials = (uint32_t*)malloc( devenum->allocated_count * sizeof(uint32_t) );
	devenum->infos = (device_information_t*)malloc( devenum->allocated_count * sizeof(device_information_t) );
	memset( devenum->infos, 0, devenum->allocated_count * sizeof(device_information_t) );

	/* Call implementation-specific directory enumerator */
	if (enumerate_devices_directory( store_device_name, devenum,
				enumerate_flags ) != result_ok)
	{
		log_debug( L"enumerate_devices_directory failed" );
		devenum->count = 0;
		return result_error;
	}

	/* Check all found devices in threads */
	if ((enumerate_flags & ENUMERATE_PROBE) && devenum->count > 0)
	{
		launch_check_threads( devenum );
	}
	log_debug( L"found %d devices", devenum->count );

	return result_ok;
}


/*
 * Exported device enumeration functions
 */

/* we do not have read/write lock so lock all DE functions with one global lock */


device_enumeration_t XIMC_API enumerate_devices(int enumerate_flags)
{
	device_enumeration_opaque_t* de;
	result_t result;
	lock_global();
	result = enumerate_devices_impl( &de, enumerate_flags );
	unlock_global();
	return result == result_ok ? (device_enumeration_t)de : 0;
}

result_t XIMC_API free_enumerate_devices(device_enumeration_t device_enumeration)
{
	device_enumeration_opaque_t* de = (device_enumeration_opaque_t*)device_enumeration;
	int i;
	lock_global();
	if (de)
	{
		for (i = 0; i < de->count; ++i)
			if (de->names[i])
				free( de->names[i] );
		if (de->serials)
			free( de->serials );
		if (de->infos)
			free( de->infos );
		/* just to be sure */
		de->names = NULL;
		de->serials = NULL;
		de->infos = NULL;
		de->count = 0;
	}
	unlock_global();
	return result_ok;
}

int XIMC_API get_device_count(device_enumeration_t device_enumeration)
{
	device_enumeration_opaque_t* de = (device_enumeration_opaque_t*)device_enumeration;
	lock_global();
	return unlocker_global( de->count );
}

pchar XIMC_API get_device_name(device_enumeration_t device_enumeration, int device_index)
{
	device_enumeration_opaque_t* de = (device_enumeration_opaque_t*)device_enumeration;
	char* device_name = NULL;
	lock_global();
	if (device_index >= 0 && device_index < de->count)
		device_name = de->names[device_index];
	unlock_global();
	return device_name;
}

result_t XIMC_API get_enumerate_device_serial(device_enumeration_t device_enumeration, int device_index, uint32_t* serial)
{
	device_enumeration_opaque_t* de = (device_enumeration_opaque_t*)device_enumeration;
	result_t result = result_ok;
	lock_global();
	if ((de->flags & ENUMERATE_PROBE) &&
			device_index >= 0 && device_index < de->count)
		*serial = de->serials[device_index];
	else
		result = result_error;
	return unlocker_global(result);
}

result_t XIMC_API get_enumerate_device_information(device_enumeration_t device_enumeration, int device_index, device_information_t* device_information)
{
	device_enumeration_opaque_t* de = (device_enumeration_opaque_t*)device_enumeration;
	result_t result = result_ok;
	lock_global();
	if ((de->flags & ENUMERATE_PROBE) &&
			device_index >= 0 && device_index < de->count)
		memcpy( device_information, &de->infos[device_index], sizeof(device_information_t) );
	else
		result = result_error;
	return unlocker_global(result);
}


result_t XIMC_API probe_device (const char* name)
{
	result_t result;
	lock_global();
	result = check_device_by_ximc_information( name, NULL, NULL ) ? result_ok : result_nodevice;
	return unlocker_global( result );
}



// vim: syntax=c tabstop=4 shiftwidth=4
