#include "common.h"
#include "ximc.h"
#include "util.h"
#include "loader.h"
#include "ximc-gen.h"
#include "protosup.h"
#include "platform.h"
#include "metadata.h"

#include "sglib.h"


typedef struct device_metadata_node_t
{
	device_t id;
	device_metadata_t data;
	struct device_metadata_node_t *next_ptr;
} device_metadata_node_t;

#define DEVICE_METADATA_COMPARATOR(e1, e2) (e1->id == e2->id ? 0 : 1)

// Global variable that holds an opened devices list
// Could be different for different processes because we are a shared library
device_metadata_node_t* g_devices_metadata = NULL;

device_metadata_t* get_metadata(device_t device)
{
	device_metadata_node_t marker, *out = NULL;
	marker.id = device;
	SGLIB_LIST_FIND_MEMBER(device_metadata_node_t, g_devices_metadata, &marker, DEVICE_METADATA_COMPARATOR, next_ptr, out);
	if (out)
		return &(out->data);
	else
		return NULL;
}

void set_metadata(device_t device, device_metadata_t metadata)
{
	device_metadata_node_t *new_dm, *out = NULL;
	new_dm = (device_metadata_node_t*)malloc( sizeof(device_metadata_node_t) );
	new_dm->id = device;
	new_dm->data = metadata;
	SGLIB_LIST_ADD_IF_NOT_MEMBER(device_metadata_node_t, g_devices_metadata, new_dm, DEVICE_METADATA_COMPARATOR, next_ptr, out);
	XIMC_UNUSED(out);
}

void remove_metadata(device_t device)
{
	device_metadata_node_t marker, *out = NULL;
	marker.id = device;
	SGLIB_LIST_DELETE_IF_MEMBER(device_metadata_node_t, g_devices_metadata, &marker, DEVICE_METADATA_COMPARATOR, next_ptr, out);
	if (out)
		free( out );
}

bool is_device_pipe (device_t device)
{
	device_metadata_t* dm = get_metadata( device );
	return dm && dm->pipe;
}

/*
 * High-level I/O
 */

// Returns 0 if synchronization succeeds
int send_synchronization_zeroes (device_t id)
{
	byte zeroes[SYNC_ZERO_COUNT];
	int res;
	int received = SYNC_ZERO_COUNT;

	log_info( L"synchronize: sending sync zeroes" );

	memset( zeroes, 0, SYNC_ZERO_COUNT );

	res = command_port_send( id, zeroes, SYNC_ZERO_COUNT );
	if (res < 0)
	{
		log_error( L"synchronize: command_port_send sync failed" );
		return result_nodevice;
	}

	for (; received > 0; --received)
	{
		res = command_port_receive( id, zeroes, 1 );
		if (res < 0)
		{
			log_error( L"synchronize: command_port_receive can't get bytes" );
			return 1;
		}
		if (zeroes[0] == 0)
		{
			log_info( L"synchronize: got a zero, done" );
			return 0;
		}
	}

	return 1;
}

result_t synchronize (device_t id)
{
	int retry_counter = SYNC_RETRY_COUNT;

	log_info( L"synchronize: started" );

	for (; retry_counter > 0; --retry_counter)
	{
		if (send_synchronization_zeroes( id ) == 0)
		{
			log_info( L"synchronize: completed" );
			return result_error;
		}
	}
	log_error( L"synchronize: synchronization attempts failed, device is lost" );
	return result_nodevice;
}

int wallclock_diff(time_t sec_beg, int msec_beg, time_t sec_end, int msec_end)
{
	int delta = 0;
	if (sec_end > sec_beg) /* be cautious */
		delta = (int)(sec_end - sec_beg);
	delta *= 1000;
	delta += msec_end - msec_beg;
	if (delta < 0)
		delta = 0;
	return delta;
}

result_t receive_synchronized (device_t id, byte* response, size_t len, bool need_sync)
{
	result_t result;
	int serial_result;
	int logical_timeout;
	int delta_time = 0;
	device_metadata_t* dm;

	time_t sec_beg, sec_cur, sec_prev;
	int msec_beg, msec_cur, msec_prev;

	dm = get_metadata( id );
	logical_timeout = dm ? dm->timeout : 0;
	if (logical_timeout <= 0)
		log_error( L"receive_synchronized: logical timeout is not properly saved at device open: %d", logical_timeout );

	get_wallclock( &sec_beg, &msec_beg );
	sec_prev = sec_beg;
	msec_prev = msec_beg;
	do
	{
		serial_result = command_port_receive( id, response, len );
		get_wallclock( &sec_cur, &msec_cur );

		switch (serial_result)
		{
			case result_serial_ok:
				return result_ok;

			case result_serial_timeout:
				log_info( L"receive_synchronized: receive timed out, requesting data from buffer one more time" );
				if (wallclock_diff( sec_prev, msec_prev, sec_cur, msec_cur ) < WAIT_BEFORE_RETRY_TIME)
				{
					log_info( L"receive_synchronized: timed out too fast, wait a little" );
					msec_sleep( WAIT_BEFORE_RETRY_TIME );
				}
				break;

			case result_serial_error:
				log_error( L"receive_synchronized: failed" );
				if (need_sync)
					return synchronize( id );
				else
					return result_error;

			case result_serial_nodevice:
				log_error( L"receive_synchronized: device lost" );
				return result_nodevice;
		}
		delta_time = wallclock_diff( sec_beg, msec_beg, sec_cur, msec_cur );
		sec_prev = sec_beg;
		msec_prev = msec_beg;
		log_info( L"receive_synchronized: passed %d msec, needed at least %d msec", delta_time, logical_timeout );
	}
	while (delta_time < logical_timeout);

	// All retries
	log_error( L"receive_synchronized: receive finally timed out" );
	if (need_sync)
	{
		if ((result = synchronize( id )) != result_ok)
		{
			log_warning( L"receive_synchronized: synchronize failed, nevermind" );
		}
	}
	else
		result = result_error;
	return result;
}

#ifdef HAVE_GIGO

handle_t open_port_gigo (const char* name, unsigned int timeout)
{
	static device_t st_gigo_device_id = 0;
	device_metadata_t dm;
	int fd;

	XIMC_UNUSED(name);
	fd = ++st_gigo_device_id;

	log_warning( L"Opening GIGO device %d...", fd );

	/* save metadata */
	dm.timeout = timeout;
	dm.pipe = false;
	dm.device_mutex = NULL;
	set_metadata( (device_t)fd, dm );

	return fd;
}

/* Garbage-In-Garbage-Out protocol */
result_t command_checked_impl (device_t id, const void* command, size_t command_len, byte* response, size_t response_len, int need_sync)
{
	XIMC_UNUSED(id);
	XIMC_UNUSED(command);
	XIMC_UNUSED(command_len);
	XIMC_UNUSED(need_sync);
	memset(response, 0, response_len);
	if (response && response_len > 4)
	{
		*((uint16_t*)(response + response_len - 2)) = get_crc(response + 4, response_len - 6);
	}
	msec_sleep( 500 + 1000 * rand() / RAND_MAX );
	return result_ok;
}

#else

result_t command_checked_impl (device_t id, const void* command, size_t command_len, byte* response, size_t response_len, int need_sync)
{
	result_t result;
	int res;
	byte errv[4] = { 'e', 'r', 'r', 'v' };

	if (command_len < 4)
		return result_error;

	if (response_len && !response)
	{
		log_error( L"command_checked can't read to empty buffer" );
	}

	// send command
	res = command_port_send( id, command, command_len );
	if (res == result_serial_error)
	{
		log_error( L"command_checked failed" );
		return result_error;
	}
	else if (res == result_serial_nodevice)
	{
		log_error( L"command_checked device lost" );
		return result_nodevice;
	}

	if (response)
	{
		// read first byte until it's non-zero
		do
		{
			if ((result = receive_synchronized( id, response, 1, need_sync )) != result_ok)
				return result;
		} while (response[0] == 0);

		// read three bytes
		if ((result = receive_synchronized( id, response+1, 3, need_sync )) != result_ok)
			return result;

		// check is it an errv answer
		if (memcmp( errv, response, (size_t)4 ) == 0)
		{
			log_warning( L"Response 'errv' received" );
			command_port_flush( id );
			return result_value_error;
		}

		// check command bytes
		if (memcmp( command, response, (size_t)4 ) != 0)
		{
			// flood the controller with zeroes
			synchronize( id );
			command_port_flush( id );
			return result_error;
		}

		// receive remaining bytes
		if ((result = receive_synchronized( id, response+4, response_len-4, need_sync )) != result_ok)
			return result;
	}

	return result_ok;
}

#endif

result_t command_checked (device_t id, const void* command, size_t command_len, byte* response, size_t response_len)
{
	return command_checked_impl( id, command, command_len, response, response_len, 1 );
}

result_t command_checked_unsynced (device_t id, const void* command, size_t command_len, byte* response, size_t response_len)
{
	return command_checked_impl( id, command, command_len, response, response_len, 0 );
}

/* Send command as a 4-byte string and return answer. Echo is checked */
result_t command_checked_str (device_t id, const char* command, byte* response, size_t response_len)
{
	return command_checked( id, command, 4, response, response_len );
}

/* same as command_checked_str but without sync */
result_t command_checked_str_unsynced (device_t id, const char* command, byte* response, size_t response_len)
{
	return command_checked_impl( id, command, 4, response, response_len, 0 );
}

/* Send command and expect exact echo command answer */
result_t command_checked_echo (device_t id, const void* command, size_t command_len)
{
	byte response[4];
	return command_checked( id, command, command_len, response, sizeof(response) );
}

/* same as command_checked_echo but without sync */
result_t command_checked_echo_unsynced (device_t id, const void* command, size_t command_len)
{
	byte response[4];
	return command_checked_impl( id, command, command_len, response, sizeof(response), 0 );
}

/* Send command as a 4-byte string and expect exact echo command answer */
result_t command_checked_echo_str (device_t id, const char* command)
{
	return command_checked_echo( id, command, 4 );
}

result_t command_checked_echo_str_unsynced (device_t id, const char* command)
{
	return command_checked_echo_unsynced( id, command, 4 );
}

result_t command_checked_echo_str_locked (device_t id, const char* command)
{
	result_t result;
	lock( id );
	result = command_checked_echo_str( id, command );
	unlock( id );
	return result;
}

/*
 * Bounds checkers
 */

result_t check_in_overrun_raw(size_t data_count, size_t buf_size)
{
	result_t result = result_ok;;
	if (data_count != buf_size)
	{
		log_error( L"buffer incoming overrun %d vs buf size %d", data_count, buf_size );
		result = result_error;
	}
	return result;
}

// layout is following:
// before PROTO14:	data(size bytes) + crc(2 bytes) + cmd(4bytes)
// PROTO14:			cmd(4bytes) + data(size bytes-4) + crc(2 bytes)
// cmd is checked at request sent
result_t check_in_overrun(device_t id, size_t data_count, size_t buf_size, const byte* response)
{
	uint16_t crc, crc_actual;
	result_t result;
	// please note: -4 is because of p+=4 shift, very hackish client code
	if ((result = check_in_overrun_raw( data_count+6-4, buf_size )) != result_ok)
		return result;
	crc = get_crc( response+4, data_count-4 );
	memcpy( &crc_actual, response+data_count, 2 );
	if (crc != crc_actual)
	{
		log_error( L"buffer crc check failed, real %x, in buffer %x", crc, crc_actual );
		// flood the controller with zeroes
		synchronize( id );
#ifndef HAVE_GIGO
		command_port_flush( id );
#endif
		return result_error;
	}
	return result_ok;
}

result_t check_in_overrun_without_crc(device_t id, size_t data_count, size_t buf_size, const byte* response)
{
	XIMC_UNUSED(id);
	XIMC_UNUSED(response);
	return check_in_overrun_raw( data_count+6-4, buf_size );
}

result_t check_out_overrun (size_t data_count, size_t buf_size)
{
	if (data_count != buf_size)
	{
		log_error( L"buffer outgoing overrun %d vs buf size %d", data_count, buf_size );
		return result_error;
	}
	return result_ok;
}

result_t check_out_atleast_overrun (size_t data_count, size_t buf_size)
{
	if (data_count > buf_size)
	{
		log_error( L"buffer outgoing overrun %d vs buf size %d", data_count, buf_size );
		return result_error;
	}
	return result_ok;
}

int powi(int x, int y)
{
	int temp;
	if (y < 0)
	{
		log_error( L"error in function powi" );
		abort();
	}
	if (y == 0)
		return 1;
	temp = powi(x, y/2);
	if (y%2 == 0)
		return temp*temp;
	else
		return x*temp*temp;
}

/*
 * Data accessors
 */
#include <byteswap.h>

void push_data(byte** where, const void* data, size_t size)
{
	memcpy( *where, data, size );
#ifdef __BIG_ENDIAN__
	if(size==2)
		** where=__bswap_16(** where);
	else if(size==4)
		** where=__bswap_32(** where);
	else if(size==8)
		** where=__bswap_64(** where);
#endif
	*where += size;
}

void push_garbage(byte** where, size_t size)
{
	memset( *where, 0xCC, size );
	*where += size;
}

// exclude four lead bytes
void push_crc (byte** where, const void* data, size_t size)
{
	uint16_t crc;
	crc = get_crc( (byte*)data + 4, size - 4 );
	memcpy( *where, &crc, 2 );
#ifdef __BIG_ENDIAN__
	** where=__bswap_16(** where);
#endif
	*where += 2;
}

// include lead bytes
void push_crc_with_command (byte** where, const void* data, size_t size)
{
	uint16_t crc;
	crc = get_crc( (byte*)data, size );
	memcpy( *where, &crc, 2 );
#ifdef __BIG_ENDIAN__
	** where=__bswap_16(** where);
#endif
	*where += 2;
}

void push_str (byte** where, const char* str)
{
	push_data( where, str, strlen( str ) );
}

void push_float (byte** where, float value)
{
	push_data( where, &value, sizeof(value) );
}

void push_double (byte** where, double value)
{
	push_data( where, &value, sizeof(value) );
}

void pop_data(byte** where, void* data, size_t size)
{
#ifdef __BIG_ENDIAN__
	if(size==2)
	{
		uint16_t tmp;
		memcpy( &tmp, *where, size );
		tmp=__bswap_16(tmp);
		memcpy( data, &tmp, size );
	}
	else if(size==4)
	{
		uint32_t tmp;
		memcpy( &tmp, *where, size );
		tmp=__bswap_32(tmp);
		memcpy( data, &tmp, size );
	}
	else if(size==8)
	{
		uint64_t tmp;
		memcpy( &tmp, *where, size );
		tmp=__bswap_64(tmp);
		memcpy( data, &tmp, size );
	}
#else
	memcpy( data, *where, size );
#endif
	*where += size;
}

void pop_str(byte** where, void* data, size_t size)
{
#ifdef __BIG_ENDIAN__
	if(size==2)
	{
		uint16_t tmp;
		memcpy( &tmp, *where, size );
		tmp=__bswap_16(tmp);
		memcpy( data, &tmp, size );
	}
	else if(size==4)
	{
		uint32_t tmp;
		memcpy( &tmp, *where, size );
		tmp=__bswap_32(tmp);
		memcpy( data, &tmp, size );
	}
	else if(size==8)
	{
		uint64_t tmp;
		memcpy( &tmp, *where, size );
		tmp=__bswap_64(tmp);
		memcpy( data, &tmp, size );
	}
#else
	memcpy( data, *where, size );
#endif
	((char*)data)[size] = '\0';
	*where += size;
}

void pop_garbage(byte** where, size_t size)
{
	*where += size;
}

float pop_float(byte** where)
{
	float result;
	memcpy( &result, *where, sizeof(result) );
	*where += sizeof(result);
#ifdef __BIG_ENDIAN__
	result=__bswap_32(*(uint32_t *)&result);
#endif
	return result;
}

double pop_double(byte** where)
{
	double result;
	memcpy( &result, *where, sizeof(result) );
	*where += sizeof(result);
#ifdef __BIG_ENDIAN__
	result=__bswap_64(*(uint64_t *)&result);
#endif
	return result;
}

#define GENERATE_PUSH(Type) \
void push_ ## Type (byte** where, Type ## _t value) \
{ \
	push_data( where, &value, sizeof(value) ); \
}

#ifdef __BIG_ENDIAN__
#define GENERATE_UPOP(Type) \
unsigned int pop_## Type (byte** where)\
{\
	Type ## _t result;\
	memcpy( &result, *where, sizeof(result) );\
	if(sizeof(result)==2)	result=__bswap_16(*(uint16_t *)&result);\
	else if(sizeof(result)==4) result=__bswap_32(*(uint32_t *)&result);\
	else if(sizeof(result)==8)	result=__bswap_64(*(uint64_t *)&result);\
	*where += sizeof(result);\
	return (unsigned int)result;\
}

#define GENERATE_IPOP(Type) \
int pop_## Type (byte** where)\
{\
	Type ## _t result;\
	memcpy( &result, *where, sizeof(result) );\
	if(sizeof(result)==2)\
		result=__bswap_16(*(uint16_t *)&result);\
	else if(sizeof(result)==4)\
		result=__bswap_32(*(uint32_t *)&result);\
	else if(sizeof(result)==8)\
		result=__bswap_64(*(uint64_t *)&result);\
	*where += sizeof(result);\
	return (int)result;\
}

/* manual pop 64-bit types definitions because default int return type is not enough */
#define GENERATE_EXACTPOP(Type) \
Type ## _t pop_## Type (byte** where)\
{\
	Type ## _t result;\
	memcpy( &result, *where, sizeof(result) );\
	if(sizeof(result)==2)\
		result=__bswap_16(*(uint16_t *)&result);\
	else if(sizeof(result)==4)\
		result=__bswap_32(*(uint32_t *)&result);\
	else if(sizeof(result)==8)\
		result=__bswap_64(*(uint64_t *)&result);\
	*where += sizeof(result);\
	return result;\
}
#else
#define GENERATE_UPOP(Type) \
unsigned int pop_## Type (byte** where)\
{\
	Type ## _t result;\
	memcpy( &result, *where, sizeof(result) );\
	*where += sizeof(result);\
	return (unsigned int)result;\
}

#define GENERATE_IPOP(Type) \
int pop_## Type (byte** where)\
{\
	Type ## _t result;\
	memcpy( &result, *where, sizeof(result) );\
	*where += sizeof(result);\
	return (int)result;\
}

/* manual pop 64-bit types definitions because default int return type is not enough */
#define GENERATE_EXACTPOP(Type) \
Type ## _t pop_## Type (byte** where)\
{\
	Type ## _t result;\
	memcpy( &result, *where, sizeof(result) );\
	*where += sizeof(result);\
	return result;\
}
#endif

GENERATE_PUSH(uint64)
GENERATE_PUSH(uint32)
GENERATE_PUSH(uint16)
GENERATE_PUSH(uint8)
GENERATE_PUSH(int64)
GENERATE_PUSH(int32)
GENERATE_PUSH(int16)
GENERATE_PUSH(int8)

GENERATE_UPOP(uint32)
GENERATE_UPOP(uint16)
GENERATE_UPOP(uint8)
GENERATE_IPOP(int32)
GENERATE_IPOP(int16)
GENERATE_IPOP(int8)

GENERATE_EXACTPOP(uint64)
GENERATE_EXACTPOP(int64)

#ifdef HAVE_LOCKS

/* 
 * Global mutex support
 */

static mutex_t* g_mutex_global = NULL;

mutex_t* mutex_global()
{
	if (!g_mutex_global)
		g_mutex_global = mutex_init( UINT_MAX );
	return g_mutex_global;
}


/* 
 * Fine-grained locks
 */

mutex_t* mutex_by_device_id(device_t id)
{
	device_metadata_t* dm;
	if (id != device_undefined)
	{
		dm = get_metadata( id );
		if (dm && dm->device_mutex)
			return dm->device_mutex;
	}
	return NULL;
}

void lock(device_t id)
{
	mutex_t* m = mutex_by_device_id(id);
	if (m)
		mutex_lock( m );
}

void unlock(device_t id)
{
	mutex_t* m = mutex_by_device_id(id);
	if (m)
		mutex_unlock( m );
}

result_t unlocker (device_t id, result_t res)
{
	mutex_t* m = mutex_by_device_id(id);
	if (m)
		mutex_unlock( m );
	return res;
}

void lock_global()
{
	mutex_lock( mutex_global() );
}

void unlock_global ()
{
	mutex_unlock( mutex_global() );
}

result_t unlocker_global (result_t res)
{
	mutex_unlock( mutex_global() );
	return res;
}

#else

void lock(device_t id)
{
	XIMC_UNUSED(id);
}

void unlock(device_t id)
{
	XIMC_UNUSED(id);
}

result_t unlocker (device_t id, result_t res)
{
	XIMC_UNUSED(id);
	return res;
}

void lock_global()
{
}

void unlock_global ()
{
}

result_t unlocker_global (result_t res)
{
	return res;
}

#endif


/*
 * Opening device
 */

device_t open_device_impl (const char* name, int timeout)
{
	handle_t fd;
	device_t device;
	device_metadata_t* dm;
	/* Port timeout is not a logical library timeout */
#ifdef HAVE_GIGO
	fd = (handle_t)open_port_gigo( name, PORT_TIMEOUT_TIME );
#else
	fd = (handle_t)open_port( name, PORT_TIMEOUT_TIME );
#endif
	if (fd == handle_invalid)
	{
		//printf("%s: ERR open port -> %d INVALID\n", __func__,fd);
		device = device_undefined;
	}
	else
	{
		device = (device_t)fd;
		/* override timeout with specified value */
		dm = get_metadata( device );
		if (dm)
		{
			dm->timeout = timeout;
#ifdef HAVE_LOCKS
			dm->device_mutex = mutex_init( (unsigned int)device );
			if (!dm->device_mutex)
			{
				if (device != device_undefined)
					close_port( (handle_t)device );
				return device_undefined;
			}
#endif
		}
	}
	return device;
}

result_t close_device_impl (device_t* id)
{
	result_t result;
	device_metadata_t* dm;
	if (*id == device_undefined)
	{
		log_error( L"attempting to close already closed device" );
		return result_error;
	}
#ifdef HAVE_LOCKS
	dm = get_metadata( *id );
	if (dm && dm->device_mutex)
		mutex_close( dm->device_mutex );
#else
	XIMC_UNUSED(dm);
#endif
	remove_metadata( *id );
#ifdef HAVE_GIGO
	result = result_ok;
#else
	result = close_port( (handle_t)*id ) == 0 ? result_ok : result_error;
#endif
	*id = device_undefined;
	return result;
}


/*
 * Exported core functions begins
 */

#if defined(__cplusplus)
extern "C" {
#endif

device_t XIMC_API open_device (const char* name)
{
	device_t device;
	lock_global();
	device = open_device_impl( name, DEFAULT_TIMEOUT_TIME );
	unlock_global();
	return device;
}

result_t XIMC_API close_device (device_t* id)
{
	result_t result;
	lock_global();
	if (id == NULL || *id == device_undefined)
		result = result_error;
	else
		result = close_device_impl( id );
	return unlocker_global( result );
}

result_t XIMC_API reset_locks()
{
	/* nop */
	return result_ok;
}

result_t XIMC_API ximc_fix_usbser_sys(const char* device_name)
{
	fix_usbser_sys( device_name );
	return result_ok;
}

result_t XIMC_API get_status (device_t id, status_t* state)
{
	lock( id );
	return unlocker( id, get_status_impl( id, state ) );
}

result_t XIMC_API get_status_calb (device_t id, status_calb_t* state, const calibration_t* calibration)
{
	lock( id );
	return unlocker( id, get_status_impl_calb( id, state, calibration ) );
}

#if defined(__cplusplus)
};
#endif

// vim: syntax=c tabstop=4 shiftwidth=4

